#include "ntddk.h"
#include "wdf.h"
#include "prototypes.h"
#include "usbdi.h"
#include "wdfusb.h"
#include "initguid.h"

DEFINE_GUID(GUID_DEVINTERFACE_WDF_SAM, // Generated using guidgen.exe
   0x573e8c73, 0xcb4, 0x4471, 0xa1, 0xbf, 0xfa, 0xb2, 0x6c, 0x31, 0xd3, 0x84);
// {573E8C73-0CB4-4471-A1BF-FAB26C31D384}

#define IOCTL_INDEX                     0x800
#define FILE_DEVICE_WDF_SAM          0x65500

#define IOCTL_WDF_SAM_SET_VANDOR_REQUEST CTL_CODE(FILE_DEVICE_WDF_SAM,\
	IOCTL_INDEX + 5, \
	METHOD_BUFFERED, \
	FILE_WRITE_ACCESS)

#define DOS_DEVICE_NAME  L"\\DosDevices\\wdf_sam"
DECLARE_CONST_UNICODE_STRING(dosDeviceName, DOS_DEVICE_NAME);

typedef struct _DEVICE_CONTEXT {
 //framework  device object  ϴ context̴.
  WDFUSBDEVICE      UsbDevice;
  WDFUSBINTERFACE   UsbInterface;
  WDFUSBPIPE                      BulkReadPipe;
  WDFUSBPIPE                      BulkWritePipe;
} DEVICE_CONTEXT, *PDEVICE_CONTEXT;

WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(DEVICE_CONTEXT, GetDeviceContext)

NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT  DriverObject,
    IN PUNICODE_STRING RegistryPath
    )
{
    WDF_DRIVER_CONFIG       config;
    NTSTATUS                status;

    KdPrint(("DriverEntry of Step2\n"));

	// PnP ̹  config->EvtDriverDeviceAdd ̺Ʈ ݹ Լ Ѵ.
    WDF_DRIVER_CONFIG_INIT(&config, EvtDeviceAdd);
	//ʿ  EvtDriverUnload ̺Ʈ ݹ  Ѵ.
	config.EvtDriverUnload =EvtDriverUnload;

	//Framework ̹ Ʈ Ѵ.
    status = WdfDriverCreate(DriverObject,
                        RegistryPath,
                        WDF_NO_OBJECT_ATTRIBUTES, 
                        &config,     
                        WDF_NO_HANDLE 
                        );

    if (!NT_SUCCESS(status)) {
        KdPrint(("WdfDriverCreate failed 0x%x\n", status));
    }

    return status;
}

NTSTATUS
EvtDeviceAdd(
    IN WDFDRIVER        Driver,
    IN PWDFDEVICE_INIT  DeviceInit
    )
{
    WDF_OBJECT_ATTRIBUTES               attributes;
    NTSTATUS                            status;
    WDFDEVICE                           device;
    PDEVICE_CONTEXT                     pDevContext;
    WDF_PNPPOWER_EVENT_CALLBACKS        pnpPowerCallbacks;
	WDF_IO_QUEUE_CONFIG                 ioQueueConfig;
    
    UNREFERENCED_PARAMETER(Driver);

	//Pnp  power ݹԼ ʱȭ  Ѵ.
    WDF_PNPPOWER_EVENT_CALLBACKS_INIT(&pnpPowerCallbacks);
    pnpPowerCallbacks.EvtDevicePrepareHardware = EvtDevicePrepareHardware;
    WdfDeviceInitSetPnpPowerEventCallbacks(DeviceInit, &pnpPowerCallbacks);

    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes, DEVICE_CONTEXT);

    status = WdfDeviceCreate(&DeviceInit, &attributes, &device);
    if (!NT_SUCCESS(status)) {
        KdPrint(("WdfDeviceCreate failed 0x%x\n", status));
        return status;
    }

	// framework device object  device context ´.
    pDevContext = GetDeviceContext(device);

	// ɺ  
	//   ɶ framework  ڵ ɺ  ش.
	status = WdfDeviceCreateSymbolicLink(device , &dosDeviceName);
	if (!NT_SUCCESS(status)) {
		KdPrint(("WdfDeviceCreateSymbolicLink failed 0x%x\n", status));
		return status;
	}

	// interface ClassGUID Ͽ ̽ ̽ Ѵ.
	//   ̽  ¿  ڵ ̽ ̽ Ȥ  Ѵ.
	status = WdfDeviceCreateDeviceInterface(device,
                                (LPGUID) &GUID_DEVINTERFACE_WDF_SAM,
                                NULL);// Reference String
    if (!NT_SUCCESS(status)) {
        KdPrint(("WdfDeviceCreateDeviceInterface failed 0x%x\n", status));
        return status;
    }

	// Deviceiocontrol ݹ ʱȭ  Ѵ.
	WDF_IO_QUEUE_CONFIG_INIT_DEFAULT_QUEUE(&ioQueueConfig,
		WdfIoQueueDispatchParallel);

	ioQueueConfig.EvtIoDeviceControl = EvtIoDeviceControl;
	ioQueueConfig.EvtIoRead = EvtIoRead;
	ioQueueConfig.EvtIoWrite = EvtIoWrite;

	status = WdfIoQueueCreate(device,
		&ioQueueConfig,
		WDF_NO_OBJECT_ATTRIBUTES,
		WDF_NO_HANDLE);

	if (!NT_SUCCESS(status)) {
		KdPrint(("WdfIoQueueCreate failed  %!STATUS!\n", status));
		return status;
	}

    return status;
}


NTSTATUS
EvtDevicePrepareHardware(
    IN WDFDEVICE    Device,
    IN WDFCMRESLIST ResourceList,
    IN WDFCMRESLIST ResourceListTranslated
    )
{
    NTSTATUS                            status;
    PDEVICE_CONTEXT                     pDeviceContext;
    WDF_USB_DEVICE_SELECT_CONFIG_PARAMS configParams;
	UCHAR i , count;

    UNREFERENCED_PARAMETER(ResourceList);
    UNREFERENCED_PARAMETER(ResourceListTranslated);

    pDeviceContext = GetDeviceContext(Device);

    //
    // Create the USB device if it is not already created.
    //
    if (pDeviceContext->UsbDevice == NULL) {
        status = WdfUsbTargetDeviceCreate(Device,
                                    WDF_NO_OBJECT_ATTRIBUTES,
                                    &pDeviceContext->UsbDevice);
        if (!NT_SUCCESS(status)) {
            KdPrint(("WdfUsbTargetDeviceCreate failed 0x%x\n", status));        
            return status;
        }
    }
    
	// USB ̽ ϱ  ʱȭ Ѵ.
    WDF_USB_DEVICE_SELECT_CONFIG_PARAMS_INIT_SINGLE_INTERFACE(&configParams);

	// ʱȭ ̽ Ѵ.
    status = WdfUsbTargetDeviceSelectConfig(pDeviceContext->UsbDevice,
                                        WDF_NO_OBJECT_ATTRIBUTES,
                                        &configParams);
    if(!NT_SUCCESS(status)) {
        KdPrint(("WdfUsbTargetDeviceSelectConfig failed 0x%x\n", status));
        return status;
    }

	//  ̽ device context Ѵ.
    pDeviceContext->UsbInterface =  
                configParams.Types.SingleInterface.ConfiguredUsbInterface;

	// pipe  Ѵ.
	count = configParams.Types.SingleInterface.NumberConfiguredPipes;

	for( i = 0 ; i< count ; i++ )
	{
		WDFUSBPIPE Pipe;
		WDF_USB_PIPE_INFORMATION pipeInfo;

		WDF_USB_PIPE_INFORMATION_INIT(&pipeInfo);

		Pipe = WdfUsbInterfaceGetConfiguredPipe(
			pDeviceContext->UsbInterface,
			i,
			&pipeInfo);// pipeInfo

		// maximumpacketsize   Ÿ   ֵ framework Ѵ.
		WdfUsbTargetPipeSetNoMaximumPacketSizeCheck(Pipe);

		if( WdfUsbPipeTypeBulk == pipeInfo.PipeType )
		{
			if( WdfUsbTargetPipeIsOutEndpoint(Pipe) )
			{
				pDeviceContext->BulkWritePipe = Pipe;
			}else{
				// WdfUsbTargetPipeIsInEndpoint(Pipe) is true
				pDeviceContext->BulkReadPipe =Pipe;
			}
		}


	}
	if( !(pDeviceContext->BulkReadPipe && pDeviceContext->BulkWritePipe) )
	{
		status = STATUS_INVALID_DEVICE_STATE;	
	}

    return status;
}

VOID
EvtDriverUnload(
	IN WDFDRIVER Driver
	)
{
	NTSTATUS status = STATUS_SUCCESS;
	KdPrint(("Wdf_sam Driver unload"));
	return ;
}


VOID 
EvtIoRead(
		  IN WDFQUEUE         Queue,
		  IN WDFREQUEST       Request,
		  IN size_t           Length
		  )   
{
	WDFUSBPIPE                  pipe;
	NTSTATUS                    status;
	WDFMEMORY                   reqMemory;
	PDEVICE_CONTEXT             pDeviceContext;
	BOOLEAN                     ret;

	UNREFERENCED_PARAMETER(Length);
	// ioqueue  framework  device object   device object context ´.
	pDeviceContext = GetDeviceContext(WdfIoQueueGetDevice(Queue));

	pipe = pDeviceContext->BulkReadPipe;

	// Read I/O request  read buffer ´.
	status = WdfRequestRetrieveOutputMemory(Request, &reqMemory);
	if(!NT_SUCCESS(status)){
		goto Exit;
	}

	// ޹ Read Request Ͽ USB pipe read request Ѵ.
	status = WdfUsbTargetPipeFormatRequestForRead(pipe,
		Request,
		reqMemory,
		NULL // Offsets NULL ΰ buffer  transfer  ȴ.
		); 
	if (!NT_SUCCESS(status)) {
		goto Exit;
	}

	// Ϸ ƾ 
	WdfRequestSetCompletionRoutine(Request,
		EvtRequestReadCompletionRoutine,
		pipe);

	// USB Io Target    USB read request Ѵ.
	ret = WdfRequestSend(Request, 
		WdfUsbTargetPipeGetIoTarget(pipe), 
		WDF_NO_SEND_OPTIONS);

	if (ret == FALSE) {
		status = WdfRequestGetStatus(Request);
		goto Exit;
	} else {
		return;
	}

Exit:
	// Read û  size information 0  ϷѴ.
	WdfRequestCompleteWithInformation(Request, status, 0);

	return;
}

VOID
EvtRequestReadCompletionRoutine(
								IN WDFREQUEST                  Request,
								IN WDFIOTARGET                 Target,
								PWDF_REQUEST_COMPLETION_PARAMS CompletionParams,
								IN WDFCONTEXT                  Context
								)
{    
	NTSTATUS    status;
	size_t      bytesRead = 0;
	PWDF_USB_REQUEST_COMPLETION_PARAMS usbCompletionParams;

	UNREFERENCED_PARAMETER(Target);
	UNREFERENCED_PARAMETER(Context);

	status = CompletionParams->IoStatus.Status;

	usbCompletionParams = CompletionParams->Parameters.Usb.Completion;

	bytesRead =  usbCompletionParams->Parameters.PipeRead.Length;

	if (NT_SUCCESS(status)){
		KdPrint(("Number of bytes read: %I64d\n", (INT64)bytesRead));  
	} else {
		KdPrint(("Read failed - request status 0x%x UsbdStatus 0x%x\n",
			status, usbCompletionParams->UsbdStatus));

	}

	// in pipe tansfer ϷǾ о Ÿ(requst  buffer) size Ѵ.
	WdfRequestCompleteWithInformation(Request, status, bytesRead);

	return;
}

VOID 
EvtIoWrite(
		   IN WDFQUEUE         Queue,
		   IN WDFREQUEST       Request,
		   IN size_t           Length
		   )   
{
	NTSTATUS                    status;
	WDFUSBPIPE                  pipe;
	WDFMEMORY                   reqMemory;
	PDEVICE_CONTEXT             pDeviceContext;
	BOOLEAN                     ret;

	UNREFERENCED_PARAMETER(Length);

	pDeviceContext = GetDeviceContext(WdfIoQueueGetDevice(Queue));

	pipe = pDeviceContext->BulkWritePipe;

	//Write I/O Input buffer ޸𸮸 ´.
	status = WdfRequestRetrieveInputMemory(Request, &reqMemory);
	if(!NT_SUCCESS(status)){
		goto Exit;
	}

	// USB Pipe  û request Ѵ.
	status = WdfUsbTargetPipeFormatRequestForWrite(pipe,
		Request,
		reqMemory,
		NULL); // Offset
	if (!NT_SUCCESS(status)) {
		goto Exit;
	}

	//Pipe write  Ϸɶ Ϸƾ Ѵ.
	WdfRequestSetCompletionRoutine(
		Request,
		EvtRequestWriteCompletionRoutine,
		pipe);
	// USB io Target  request Ѵ.
	ret = WdfRequestSend(Request, 
		WdfUsbTargetPipeGetIoTarget(pipe), 
		WDF_NO_SEND_OPTIONS);
	if (ret == FALSE) {
		status = WdfRequestGetStatus(Request);
		goto Exit;
	} else {
		return;
	}

Exit:    
	// USB I/O target  request  size information 0 ϷŲ.
	WdfRequestCompleteWithInformation(Request, status, 0);

	return;
}

VOID
EvtRequestWriteCompletionRoutine(
								 IN WDFREQUEST                  Request,
								 IN WDFIOTARGET                 Target,
								 PWDF_REQUEST_COMPLETION_PARAMS CompletionParams,
								 IN WDFCONTEXT                  Context
								 )
{
	NTSTATUS    status;
	size_t      bytesWritten = 0;
	PWDF_USB_REQUEST_COMPLETION_PARAMS usbCompletionParams;

	UNREFERENCED_PARAMETER(Target);
	UNREFERENCED_PARAMETER(Context);

	status = CompletionParams->IoStatus.Status;

	usbCompletionParams = CompletionParams->Parameters.Usb.Completion;

	bytesWritten =  usbCompletionParams->Parameters.PipeWrite.Length;

	if (NT_SUCCESS(status)){
		KdPrint(("Number of bytes written: %I64d\n", (INT64)bytesWritten));        
	} else {
		KdPrint(("Write failed: request Status 0x%x UsbdStatus 0x%x\n", 
			status, usbCompletionParams->UsbdStatus));
	}

	WdfRequestCompleteWithInformation(Request, status, bytesWritten);

	return;
}


VOID
EvtIoDeviceControl(
				   IN WDFQUEUE   Queue,
				   IN WDFREQUEST Request,
				   IN size_t     OutputBufferLength,
				   IN size_t     InputBufferLength,
				   IN ULONG      IoControlCode    
				   )
{
	WDFDEVICE                           device;
	PDEVICE_CONTEXT                     pDevContext;
	size_t                              bytesTransferred = 0;
	NTSTATUS                            status;
	WDF_USB_CONTROL_SETUP_PACKET        controlSetupPacket;
	WDF_MEMORY_DESCRIPTOR               memDesc;
	WDFMEMORY                           memory;

	UNREFERENCED_PARAMETER(InputBufferLength);
	UNREFERENCED_PARAMETER(OutputBufferLength);

	device = WdfIoQueueGetDevice(Queue);
	pDevContext = GetDeviceContext(device);

	switch(IoControlCode) {

	case IOCTL_WDF_SAM_SET_VANDOR_REQUEST:

		if(InputBufferLength < sizeof(UCHAR)) {
			status = STATUS_BUFFER_OVERFLOW;
			bytesTransferred = sizeof(UCHAR);
			break;
		} 

		//DeviceIoControl request Inpubuffer  ´.
		status = WdfRequestRetrieveInputMemory(Request, &memory);
		if (!NT_SUCCESS(status)) {
			KdPrint(("WdfRequestRetrieveMemory failed 0x%x", status));
			break;
		}

		// Conrol Setup packet Ѵ.
		WDF_USB_CONTROL_SETUP_PACKET_INIT_VENDOR(&controlSetupPacket,
			BmRequestHostToDevice,
			BmRequestToDevice,
			0, // Request
			0, // Value
			0); // Index                                                        

		// request ޵ wdfmemory WDF Memory descriptor Ѵ.
		WDF_MEMORY_DESCRIPTOR_INIT_HANDLE(&memDesc, memory, NULL);

		// ũ  control transfer Ѵ.
		status = WdfUsbTargetDeviceSendControlTransferSynchronously(
			pDevContext->UsbDevice, 
			NULL, // Optional WDFREQUEST
			NULL, // PWDF_REQUEST_SEND_OPTIONS
			&controlSetupPacket,
			&memDesc,
			(PULONG)&bytesTransferred);
		if (!NT_SUCCESS(status)) {
			KdPrint(("SendControlTransfer failed 0x%x", status));
			break;
		}
		break;

	default:
		status = STATUS_INVALID_DEVICE_REQUEST;
		break;
	}

	WdfRequestCompleteWithInformation(Request, status, bytesTransferred);

	return;
}
